const byte tckPin = 3;	// D0
const byte tdiPin = 2;	// D1
const byte tdoPin = 0;	// D2
const byte tmsPin = 1;	// D3

const byte tckBit = 1;
const byte tdiBit = 2;
const byte tdoBit = 4;
const byte tmsBit = 8;

#define BUFLEN 62

byte tdobuf[BUFLEN];
byte tdibuf[BUFLEN];
char cmdbuf[10];

void setup() 
{
	pinMode(tckPin, OUTPUT);
	pinMode(tdiPin, OUTPUT);
	pinMode(tmsPin, OUTPUT);
	pinMode(tdoPin, INPUT);
	Serial.begin(115200);
	Serial.setTimeout(2000);
}

void tap_reset()
{
	int i;
	PORTD |= tmsBit;
	for (i=0; i<5; i++) {
		PORTD |= tckBit;
		PORTD &= ~tckBit;
	}
	PORTD &= ~tmsBit;
}

void tck_pulse(int n)
{
	while(n--) {
		PORTD |= tckBit;
		PORTD &= ~tckBit;
	}
}

inline boolean tck_pulse_in(boolean tdidata)
{
	boolean ret = (PIND & tdoBit) == tdoBit;
	if(tdidata)
		PORTD |= tdiBit;
	else
		PORTD &= ~tdiBit;
	PORTD |= tckBit;
	PORTD &= ~tckBit;
	
	return ret;
}

inline void tap_move(int tmsval)
{
	digitalWrite(tmsPin, tmsval);
	tck_pulse(1);
}

void read_idcode()
{
	int i, j;
	byte id;
	tap_reset();	// TLR
	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	tap_move(0);	// CAPTURE DR
	tap_move(0);	// SHIFT DR
	for(i=0; i<3; i++) {
		for(j=0; j<8; j++) {
			id >>= 1;
			if(tck_pulse_in(true))
				id |= 0x80;
		}
		tdobuf[i] = id;
	}
	for(j=0; j<7; j++) {
		id >>= 1;
		if(tck_pulse_in(true))
			id |= 0x80;
	}
	id >>= 1;
	PORTD |= tmsBit;	// EXIT DR
	if(tck_pulse_in(true))
		id |= 0x80;
	tdobuf[3] = id;
	tap_move(1);	// UPDATE DR
	tap_move(0);	// RTI
}

void shift_loop(word nloop, byte rem, boolean tdimode)
{
	int i,j;
	byte indata, outdata;
	
	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	if(tdimode) tap_move(1);	// SELECT IR
	tap_move(0);	// CAPTURE DR/IR
	tap_move(0);	// SHIFT DR/IR
	for(i=0; i<nloop; i++) {
		indata = tdibuf[i];
		for(j=0; j<8; j++) {
			outdata >>= 1;
			if(tck_pulse_in(indata&0x01)) {
				outdata |= 0x80;
			}
			indata >>= 1;
		}
		tdobuf[i] = outdata;
	}
	indata = tdibuf[nloop];
	for(j=0; j<rem-1; j++) {
		outdata >>= 1;
		if(tck_pulse_in(indata&0x01)) {
			outdata |= 0x80;
		}
		indata >>= 1;
	}
	outdata >>= 1;
	PORTD |= tmsBit;	// EXIT DR/IR
	if(tck_pulse_in(indata&0x01))
		outdata |= 0x80;
	tdobuf[nloop] = outdata;
	tap_move(1);	// UPDATE DR/IR
	tap_move(0);	// RTI
}

byte xilinx_sir(byte cmd)
{
	int i;
	byte outdata;
	
	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	tap_move(1);	// SELECT IR
	tap_move(0);	// CAPTURE IR
	tap_move(0);	// SHIFT IR
	for(i=0; i<5; i++) {
		outdata >>= 1;
		if(tck_pulse_in(cmd&0x01))
			outdata |= 0x80;
		cmd >>= 1;
	}
	outdata >>= 1;
	PORTD |= tmsBit;	// EXIT IR
	if(tck_pulse_in(cmd&0x01))
		outdata |= 0x80;
	tap_move(1);	// UPDATE IR
	tap_move(0);	// RTI
	return outdata;
}

byte xilinx_sdr(byte cmd)
{
	int i;
	byte outdata;
	
	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	tap_move(0);	// CAPTURE DR
	tap_move(0);	// SHIFT DR
	for(i=0; i<8; i++) {
		outdata >>= 1;
		if(tck_pulse_in(cmd&0x01))
			outdata |= 0x80;
		cmd >>= 1;
	}
	outdata >>= 1;
	PORTD |= tmsBit;	// EXIT DR
	if(tck_pulse_in(cmd&0x01))
		outdata |= 0x80;
	tap_move(1);	// UPDATE DR
	tap_move(0);	// RTI
	return outdata;
}

void get_idcode()
{
	int i;
	read_idcode();
	for(i=0; i<4; i++) {
		Serial.write(tdobuf[i]);
	}
}

void config_fpga1()
{
	memset(tdibuf, 0, 32);
	tap_reset();
	xilinx_sir(0x3f);
	xilinx_sir(0x0b);
	xilinx_sir(0x05);
	tck_pulse(14000);
	xilinx_sir(0x05);
	shift_loop(3, 8, false);
	xilinx_sir(0x05);
}

void send_bitdata()
{
	int i,timeout;
	byte indata, j, k;

	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	tap_move(0);	// CAPTURE DR
	tap_move(0);	// SHIFT DR

	for(i=0; i<2729; i++) {
		timeout = 0;
		while(Serial.available() < 62){
			if(timeout++>30000) {
				Serial.write('Q');
				tap_reset();
				return;
			}
		}
		Serial.readBytes((char*)tdibuf, 62);
		for(j=0; j<62; j++) {
			indata = tdibuf[j];
			for(k=0; k<8; k++) {
				tck_pulse_in(indata&0x80);
				indata <<= 1;
			}
		}
		Serial.write('B');
	}
	Serial.readBytes((char*)tdibuf, 18);
	for(j=0; j<17; j++) {
		indata = tdibuf[j];
		for(k=0; k<8; k++) {
			tck_pulse_in(indata&0x80);
			indata <<= 1;
		}
	}

	indata = tdibuf[17];
	for(j=0; j<7; j++) {
		tck_pulse_in(indata&0x80);
		indata <<= 1;
	}
	PORTD |= tmsBit;	// EXIT DR
	tck_pulse_in(indata&0x80);
}			

boolean config_fpga2()
{
	byte done;
	xilinx_sir(0x0c);
	tck_pulse(12);
	xilinx_sir(0x3f);
	xilinx_sir(0x3f);
	xilinx_sir(0x0c);
	tck_pulse(12);
//	delay(1000);
	done = xilinx_sir(0x3f);
	xilinx_sir(0x3f);
	tdibuf[0] = 0;
	shift_loop(0, 1, false);
	return (done & 0x80) == 0x80;
}

void start_capture()
{
	xilinx_sir(0x03);
	tdibuf[0] |= 0x01;	// RESET
	shift_loop(2, 8, false);
	tdibuf[0] &= ~0x01;	// release RESET
	shift_loop(2, 8, false);
}

void get_waveform()
{	
	int i;
	byte mem, j, k;

	xilinx_sir(0x02);	// send USER1

	tap_move(0);	// RTI
	tap_move(1);	// SELECT DR
	tap_move(0);	// CAPTURE DR
	tap_move(0);	// SHIFT DR

	for(i=0; i<396; i++) {
		for(j=0; j<62; j++) {
			for(k=0; k<8; k++) {
				mem >>= 1;
				if(tck_pulse_in(0))
					mem |= 0x80;
			}
			Serial.write(mem);
		}
	}
	for(j=0; j<24; j++) {
		for(k=0; k<8; k++) {
			mem >>= 1;
			if(tck_pulse_in(0))
				mem |= 0x80;
		}
		Serial.write(mem);
	}
	for(j=0; j<7; j++) {
		mem >>= 1;
		if(tck_pulse_in(0))
			mem |= 0x80;
	}
	mem >>= 1;
	PORTD |= tmsBit;	// EXIT DR
	if(tck_pulse_in(0))
		mem |= 0x80;
	tap_move(1);	// UPDATE DR
	tap_move(0);	// RTI
	Serial.write(mem);
	Serial.print("END");
}

void get_status()
{
	int i;
	xilinx_sir(0x03);
	shift_loop(6, 8, false);	// final 4, 8
	for(i=0; i<7; i++) {	// final 5
		Serial.write(tdobuf[i]);
	}
}

void loop() {
	int nloop, i, nbytes;
	byte rem;

	while(true) {
		nbytes = Serial.available();
		if(nbytes >= 7) {
			for(i=0; i<nbytes; i++) {
				if(Serial.peek() == '#') {
					if(nbytes-i >= 7) {
						Serial.readBytes(cmdbuf, 7);
					}
					break;
				}
				else {
					Serial.read();
				}
			}
			cmdbuf[7] = '\0';
			if(!strcmp(cmdbuf, "#HELLO!")) {
				Serial.print("DaVinci");
			}
			else if(!strcmp(cmdbuf, "#IDCODE")) {
				get_idcode();
			}
			else if(!strcmp(cmdbuf, "#CONFIG")) {
				config_fpga1();
				Serial.print("CONFIG");
				send_bitdata();
				if(config_fpga2())
					Serial.print("OK");
				else
					Serial.print("NG");
			}
			else if(!strcmp(cmdbuf, "#CAPTUR")) {
				Serial.readBytes((char*)tdibuf, 3);
				start_capture();
			}
			else if(!strcmp(cmdbuf, "#MEMORY")) {
				get_waveform();
			}
			else if(!strcmp(cmdbuf, "#STATUS")) {
				Serial.readBytes((char*)tdibuf, 7);	// final = 5
				get_status();
			}
			else {
				while(Serial.read() >= 0) {;}
			}
			cmdbuf[0] = '\0';
		}
	}
}
